#!/usr/bin/env python
# coding: utf-8

import argparse
import configparser
import re
import pathlib as pl

import datetime as dt

from spifpy.input import DMTMonoFile, DMTGreyFile, SPECFile, TwoDFile
from spifpy.spif import SPIFCore

inst_dict = {'2DC': TwoDFile,
             '2DP': TwoDFile,
             'CIP': DMTMonoFile,
             'CIPGS': DMTGreyFile,
             'PIP': DMTMonoFile,
             '2DS': SPECFile,
             'HVPS': SPECFile}

def extract():

     # Get the parser and grab the args

     parser = get_parser()
     args = parser.parse_args()

     # Check the args, make sure everything is OK before using
     # the arguments to do anything with them

     args_checker = ArgsChecker(args)
     args_checker.check_args()

     # Do some more detailed checking of the times

     time_checker = SPIFStartEndTimeChecker(args.filename, args.config)
     time_checker.check_time_args(args.start, args.end)

     # Transform the arguments to the proper forms for smooth processing

     args_transformer = ArgsTransformer(args)
     transformed_args = args_transformer.transform_args()

     spif_core = call_spifcore(transformed_args)
     spif_core.process(processors=transformed_args['nproc'])

def get_parser():
    parser = argparse.ArgumentParser(description='Processes raw OAP data to' +
                                                 ' SPIF formatted NetCDF file')
    parser.add_argument('filename',
                        type=str,
                        help= 'path to raw instrument file to process.')

    parser.add_argument('config',
                        type=str,
                        help= 'path to config file to use for processing')

    parser.add_argument('-o',
                        dest='output',
                        type=str,
                        help='Filename to use for SPIF output',
                        default=None)

    parser.add_argument('--start',
                        dest='start',
                        type=str,
                        help='The start time in the file to begin processing.',
                        default=None)

    parser.add_argument('--end',
                        dest='end',
                        type=str,
                        help='The end time in the file to begin processing.',
                        default=None)

    parser.add_argument('-n',
                        dest='nproc',
                        type=int,
                        help='number of processors to use in processing',
                        default=None)

    return parser

def call_spifcore(transformed_args):
     inst_name = get_inst_name(transformed_args)
     filename = transformed_args['filename']
     outfile = transformed_args['output']
     config = transformed_args['config']
     start_time = transformed_args['start']
     end_time = transformed_args['end']

     spif_core = SPIFCore(
          inst_dict[inst_name],
          filename,
          outfile,
          config,
          start_time = start_time,
          end_time = end_time
     )

     return spif_core

def get_inst_name(transformed_args):
     config = configparser.ConfigParser(allow_no_value=True)

     config.read(transformed_args['config'])
     inst_name = config['instrument'].get('instrument_name', None)

     return inst_name

class ArgsChecker:
     def __init__(self, args):
          self.args = args

     def check_args(self):
          self.check_filename(self.args.filename)
          self.check_config(self.args.config)

          if self.args.output is not None : self.check_output(self.args.output)
          if self.args.start is not None : self.check_time_args(self.args.start)
          if self.args.end is not None : self.check_time_args(self.args.end)
          if self.args.nproc is not None : self.check_nproc(self.args.nproc)

     def check_filename(self, filename):

          try:
               assert pl.Path(filename).is_file()
          except AssertionError:
               print(f"ERROR : The file {filename} is not a valid file")
               raise

          try:
               assert pl.Path(filename).stat().st_size
          except AssertionError:
               print(f"ERROR : The file {filename} is an empty file")
               raise

     def check_config(self, config_file):
          config = configparser.ConfigParser(allow_no_value=True)

          config.read(config_file)
          inst_name = config['instrument'].get('instrument_name', None)

          try:
               assert inst_name in inst_dict
          except AssertionError:
               print(f"ERROR : The provided instrument type {inst_name} from config file {config} is invalid. Please provide a valid instrument from the following list {' | '.join([k for k in inst_dict.keys()])}")
               raise

     def check_output(self, output):

          try:
               open(output, 'w')
          except OSError:
               print(f"ERROR : The provided output filename {output} is not a valid filename.")
               raise

          try:
               assert pl.Path(output).suffix == '.nc'
          except AssertionError:
               print(f"ERROR : The provided filename {output} does not have a .nc file ending")
               raise

     def check_time_args(self, time):
          try:
               assert re.match('[0-9]{2}:[0-9]{2}:[0-9]{2}', time)
          except AssertionError:
               print(f"ERROR : The provided time {time} does not match the required format of HH:MM:SS")

          split_time = [int(x) for x in time.split(':')]

          try:
               assert split_time[0] < 24 and split_time[0] >= 0
               assert split_time[1] < 60 and split_time[0] >= 0
               assert split_time[2] < 60 and split_time[0] >= 0
          except AssertionError:
               print(f"ERROR : The provided time {time} does not have proper digits")
               raise

     def check_nproc(self, nproc):
          try:
               assert isinstance(nproc, int)
          except AssertionError:
               print(f"ERROR : The provided argument for nproc {nproc} is not a number.")
               raise

          try:
               assert nproc > 0
          except AssertionError:
               print(f"ERROR : The provided argument nproc {nproc} is less than one")
               raise

class SPIFStartEndTimeChecker:

     def __init__(self, filename, config_file):
          self.filename = filename
          self.config_file = config_file

          self.start_time = None
          self.end_time = None

          self.get_start_end_time()

     def get_start_end_time(self):
          config = configparser.ConfigParser(allow_no_value=True)

          config.read(self.config_file)
          inst_name = config['instrument'].get('instrument_name', None)

          inst_class = inst_dict[inst_name]

          self.inst_file = inst_class(
               self.filename,
               config['instrument']['instrument_name'],
               config['resolution']['value']
          )

          self.inst_file.read()

          file_datetimes = self.inst_file.datetimes

          self.file_time_start = self.inst_file.start_date - file_datetimes[0].replace(microsecond = 0)
          self.file_time_end = file_datetimes[-1].replace(microsecond = 0) - self.inst_file.start_date

     def check_time_args(self, start_time, end_time):
          if start_time is not None:
               start_time_sec = [int(x) for x in start_time.split(':')]
               start_time_sec = dt.timedelta(
                    seconds = start_time_sec[0]*3600 + start_time_sec[1]*60 + start_time_sec[2]
               )

          if end_time is not None:
               end_time_sec = [int(x) for x in end_time.split(':')]
               end_time_sec = dt.timedelta(
                    seconds = end_time_sec[0]*3600 + end_time_sec[1]*60 + end_time_sec[2]
               )

          if (start_time is not None) and (end_time is not None):
               try:
                    assert start_time_sec < end_time_sec
               except AssertionError:
                    print(f"ERROR : Start time {start_time} is greater than end time {end_time}")
                    raise

          if start_time is not None:
               try:
                    self.check_time_in_file(start_time_sec)
               except AssertionError:
                    print(f"ERROR : Start time {start_time} is not in the file time range")
                    raise

          if end_time is not None:
               try:
                    self.check_time_in_file(end_time_sec)
               except AssertionError:
                    print(f"ERROR : End time {end_time} is not in the file time range")
                    raise

     def check_time_in_file(self, time):
          assert time >= self.file_time_start
          assert time <= self.file_time_end

class ArgsTransformer:

     def __init__(self, args) -> None:
         self.args = args

         self.transformed_args = {
              'filename':None,
              'config':None,
              'output':None,
              'start':None,
              'end':None,
              'nproc':None,
              'aux_file':None,
              'aux_config':None,
         }

     def transform_args(self):
          self.transform_filename()
          self.transform_config()
          self.transform_output()
          self.transform_start()
          self.transform_end()
          self.transform_nproc()

          return self.transformed_args

     def transform_filename(self):
          self.transformed_args['filename'] = pl.Path(self.args.filename)

     def transform_config(self):
          self.transformed_args['config'] = self.args.config

     def transform_output(self):
          if self.args.output == None:
               input = pl.Path(self.args.filename)
               self.transformed_args['output'] = input.parent / (input.name.replace('.','_') + '.nc')
          else:
               self.transformed_args['output'] = pl.Path(self.args.output)

     def transform_start(self):
          if self.args.start is None:
               self.transformed_args['start'] = None
          else:
               start = [int(x) for x in self.args.start.split(':')]
               start = start[0]*3600 + start[1]*60 + start[2]
               self.transformed_args['start'] = dt.timedelta(seconds = start)

     def transform_end(self):
          if self.args.end is None:
               self.transformed_args['end'] = None
          else:
               end = [int(x) for x in self.args.end.split(':')]
               end = end[0]*3600 + end[1]*60 + end[2]
               self.transformed_args['end'] = dt.timedelta(seconds = end)

     def transform_nproc(self):
          self.transformed_args['nproc'] = self.args.nproc



"""A module to handle reading SEA Model 200 (M200) files.

This module will read data from an M200 file

Before reading through this code, it is highly recommended that you become
familiar with a few other topics. First, read Chapter 3 of the SEA Model 200
User's Manual (http://www.scieng.com/pdf/m200.pdf); it describes the data
format used by the M200, and by extension, the format of the data that this
program reads. Second, become familiar with Python's 'struct' module
(http://docs.python.org/2/library/struct.html); it effectively allows a
program to read in blocks of data from a file as though they were C-style
structs. Finally, Python's list comprehensions
(http://docs.python.org/2/tutorial/datastructures.html#list-comprehensions)
are used frequently when reading a file.

"""
# TODO update the above docstring

import collections
import csv
import datetime
import math
import re
import struct
import warnings

from tqdm import tqdm

from spifpy.input import tag

M200_ACQUISITION_TABLE = 'ACQTBL.TXT'
M200_SECONDARY_ACQUISITION_TABLE = 'ACQ2TBL.TXT'
M300_ACQUISITION_TABLE = 'acq.300'
M300_SECONDARY_ACQUISITION_TABLE = 'saq.300'

# These are the tag numbers that are defined in the user manuals of the
# SEA Model 200 and SEA Model 300.
TIME_TAG = 0
NEXT_TAG = 999
RESERVED_TAG_LOW_END = 65000
RESERVED_TAG_HIGH_END = 65529
FILENAME_TAG = 65530
FILEDATA_TAG = 65531
COMMAND_TAG = 65532
ERROR_TAG = 65533
SAME_TAG = 65534
LAST_TAG = 65535

# These acquisition types are defined in the user manuals of the
# SEA Model 200 and SEA Model 300.
SEA_ACQUISITION_TYPE_2D_MONO_IMAGE = 5
SEA_ACQUISITION_TYPE_SEA_ANALOG_TO_DIGITAL_INPUT = 35
SEA_ACQUISITION_TYPE_2D_GREY_ADVANCED = 66
SEA_ACQUISITION_TYPE_CIP_IMAGE = 78
SEA_ACQUISITION_TYPE_NETWORK_BINARY_DATA = 85

# This struct represents the format used to store time data in the data files.
# Using this struct, the time data can be easily read from the data files.
SEA_TIME_STRUCT = struct.Struct(''.join(('<', 'H' * 9)))

# This named tuple will be used to store the components of the time data from an
# SEA file in a convenient, simple data structure.
SEATime = collections.namedtuple('SEATime',
    'year, month, day, hour, minute, second, fraction_of_second, max_sys_freq, '
    'buffer_life_span')


class SEAReadWarning(UserWarning):
    """A warning issued when unexpected data is encountered. """
    pass


class InvalidSEAFileException(Exception):
    """Raised when an invalid SEA file is passed to M200File."""
    def __init__(self, filename):
        self.filename = filename


class IncompleteDirectoryEntryException(Exception):
    """Raised when an incomplete directory entry is encountered when readig data
    from an SEA file.

    """
    pass


class UnexpectedEndOfFileException(Exception):
    """Raised when the end of a file is reached in the middle of reading a
    record.

    """
    pass


class UnexpectedEndOfBufferException(Exception):
    """Raised when the end of a buffer is reached in the middle of reading a
    record.

    """
    pass


def datetime_from_raw_seatime(sea_file_data, offset=0):
    """Parses time data from an SEA dataset return a datetime object containing
    the data.

    Arguments:
    sea_file_data -- A buffer that contains SEA data
    offset -- The offset into sea_file_data where the time data is found

    """
    seatime = SEATime(*SEA_TIME_STRUCT.unpack_from(sea_file_data, offset))

    return datetime_from_seatime(seatime)


def datetime_from_seatime(seatime):
    # - max_sys_freq is the number of clock ticks per second
    # - fraction_of_second is the number of clock ticks that have occurred in
    #   the current second
    # - multiply by 1 000 000 to convert fraction of second into microseconds
    #   (needed by datetime)
    microsecond = int(1000000 * seatime.fraction_of_second / seatime.max_sys_freq)

    try:
        datetime_obj = datetime.datetime(seatime.year, seatime.month, seatime.day,
             seatime.hour, seatime.minute, seatime.second, microsecond,
             tzinfo=None)
    except ValueError as ve:
        print(seatime)
        raise ve

    return datetime_obj


csv.register_dialect('sea_acquisition_table', delimiter=' ', quotechar='"',
                        skipinitialspace=True)


class SEAFile(object):
    """A class dedicated to reading data from a file generated by the M200.

    Public Methods:
        get_datetime -- returns a datetime object that contains the start time
            of the file
        get_tag_by_number -- returns the tag from this file that has the
            specified tag number
        get_tags_by_typ -- returns a list of all tags from this file that have
            the specified type
        iter_buffers -- return a generator that will return each buffer of this
            file in sequence

    Public Attributes:
        n/a

    """
    #TODO fix docstring above

    _ACQ_NAME = None
    _ACQ_CSV_DIALECT = 'sea_acquisition_table'
    _ACQ2_NAME = None
    _TAG_CLASS = tag.SEATag
    _TIME_DATASET_NUMBER_OF_SAMPLES = None
    _TIME_DATASET_BYTES_PER_SAMPLE = 18
    _TIME_DATASET_ADDRESS = 0xAA55

    def __init__(self, filename):
        self.filename = filename

        self._already_initialized = False
        self._buffer_types = set()
        self._config_files = []
        self._datetime = None
        self._file_contents = None
        self._secondary_tags = set()
        self._tags_by_tag_number = {}
        self._tags_by_typ = {}

        self._filetype_check()

        self._init_from_file()

    def _filetype_check(self):
        """Checks that the underlying file is a valid SEA file.

        This method does not actually verify the entire file. Instead, it
        attempts to read the first directory entry of the first directory in the
        file. This directory entry will always be for a time dataset, so we can
        compare known values for time directories to see if they match.

        """
        try:
            dir_entry = DirectoryEntry.new_from_raw_data(self._get_contents())
        except IncompleteDirectoryEntryException:
            raise InvalidSEAFileException(self.filename)

        if dir_entry.tag_number != TIME_TAG or \
                dir_entry.bytes_per_sample != \
                    self._TIME_DATASET_BYTES_PER_SAMPLE or \
                dir_entry.address != self._TIME_DATASET_ADDRESS:
            raise InvalidSEAFileException(self.filename)

    def _add_tag(self, tag):
        if tag.tag_number in self._tags_by_tag_number.keys():
            warnings.warn(
                '{:s}.add_tag: ignoring duplicate tag '
                '({:s})'.format(self.__class__.__name__, tag))
            return

        self._tags_by_tag_number[tag.tag_number] = tag
        try:
            self._tags_by_typ[tag.typ].append(tag)
        except KeyError as ke:
            if ke.args[0] == tag.typ:
                self._tags_by_typ[tag.typ] = [tag]
            else:
                raise ke

    def get_datetime(self):
        return self._datetime

    def get_tag_by_tag_number(self, tag_number):
        """Fetches a tag from self.

        Arguments:
            tag_number -- The number of the tag to retrieve

        Returns:
            An instance of Tag, owned by self, which has the given tag number

        Raises:
            KeyError -- Raised when self has no tag with the given number

        """
        return self._tags_by_tag_number[tag_number]

    def get_tags_by_typ(self, typ):
        """
        Fetches a group of tags with a given typ.

        Note: 'typ' is not a typo. It is used intentionally to avoid name
        conflicts with the type keyword/class.

        Arguments:
            typ -- The typ value of tags to be fetched.

        Returns:
            A list of all Tag instances, owned by self, which have the given typ
            attribute.

        Raises:
            KeyError -- Raised if self has no Tags of with the given typ(e).

        """

        return self._tags_by_typ[typ]

    def _get_contents(self):
        """Fetches the contents of the SEA file that self represents.

        Returns:
            A long string which holds the contents of the represented SEA file.

        Implementation Details:
            The first time this method is called, it will read in the data,
            store it in an attribute, and then return the data. On subsequent
            calls, this method simply returns the attribute.

        """

        if self._file_contents is None:
            with open(self.filename, mode='rb') as data_file:
                self._file_contents = data_file.read()

        return self._file_contents

    def _init_from_file(self):
        """Initializes self using the contents of the SEA file that self
        represents.

        This method skims through the SEA directories in the SEA file and
        reads those that contain metadata that will be needed to read the rest
        of the file. The most important aspect is reading the FILEDATA data
        containing the acquisition table, which contains descriptions and tag
        numbers for each data type contained in the file. In addition, the
        acquisition table specifies how the various data types are grouped into
        buffers.

        Arguments:
            n/a

        Returns:
            n/a

        """
        if self._already_initialized:
            return

        for dir_entries, raw_buffer in \
                self._raw_buffers({'tag_number': [FILEDATA_TAG]}):
            has_filename = False
            for dir_entry in dir_entries:
                # print(dir_entry.tag_number)
                if dir_entry.tag_number == TIME_TAG:
                    # Set a date-time value for the file as a whole, using the
                    # first time data found in the file.
                    if self._datetime is None:
                        self._datetime = datetime_from_raw_seatime(
                            raw_buffer, dir_entry.data_offset)

                elif dir_entry.tag_number == NEXT_TAG:
                    break

                elif RESERVED_TAG_LOW_END <= dir_entry.tag_number <= \
                        RESERVED_TAG_HIGH_END:
                    # reserved for future use
                    pass

                elif dir_entry.tag_number == FILENAME_TAG:
                    filename = self._unpack_string_data(
                            raw_buffer, dir_entry
                            )[0].strip(b'\0').decode('utf-8')

                    new_file = ConfigFile(filename)
                    has_filename = True

                elif dir_entry.tag_number == FILEDATA_TAG:
                    if has_filename:
                        new_file.data = self._unpack_string_data(
                            raw_buffer, dir_entry
                            )[0].strip(b'\n')

                        self._config_files.append(new_file)
                        has_filename = False

                        if new_file.name == M200_ACQUISITION_TABLE:
                            # If the tag class is still the "generic" SEATag,
                            # change it to the more specific version
                            if self._TAG_CLASS == tag.SEATag:
                                self._TAG_CLASS = tag.M200Tag
                                self._TIME_DATASET_NUMBER_OF_SAMPLES = 1
                                self._add_tag(self._TAG_CLASS('Time', -1,
                                    TIME_TAG,
                                    self._TIME_DATASET_NUMBER_OF_SAMPLES,
                                    self._TIME_DATASET_BYTES_PER_SAMPLE, 0, 0,
                                    0, 0, 0xAA55))
                                self._add_tag(self._TAG_CLASS('Filename', -1,
                                    FILENAME_TAG, 1, 0, 255, 0, 0, 0,
                                    0xAA55))
                                self._add_tag(self._TAG_CLASS('Filedata', -1,
                                    FILEDATA_TAG, 1, 0, 255, 0, 0, 0,
                                    0xAA55))

                            self._parse_acqtbl(new_file)

                        elif new_file.name == M300_ACQUISITION_TABLE:
                            # If the tag class is still the "generic" SEATag,
                            # change it to the more specific version

                            if self._TAG_CLASS == tag.SEATag:
                                self._TAG_CLASS = tag.M300Tag
                                self._TIME_DATASET_NUMBER_OF_SAMPLES = 2
                                self._add_tag(self._TAG_CLASS('Time', TIME_TAG,
                                    self._TIME_DATASET_NUMBER_OF_SAMPLES, 1,
                                    self._TIME_DATASET_BYTES_PER_SAMPLE, 0, 0,
                                    0, 0, 'DummyTimeBoard', 0))
                                self._add_tag(self._TAG_CLASS("Filename",
                                    FILENAME_TAG, 1, 1, 0, 255, 0, 0, 0,
                                    'DummyFilenameBoard'))
                                self._add_tag(self._TAG_CLASS("Filedata",
                                    FILEDATA_TAG, 1, 1, 0, 255, 0, 0, 0,
                                    'DummyFiledataBoard'))

                            self._parse_acqtbl(new_file)

                        # TODO actually secondary acquisition data
                        elif new_file.name == M200_SECONDARY_ACQUISITION_TABLE or \
                                new_file.name == M300_SECONDARY_ACQUISITION_TABLE:
                            self._parse_acq2tbl(new_file)

                    else:
                        print('Found FILEDATA but no associated FILENAME.')

                elif dir_entry.tag_number == LAST_TAG:
                    break

        # We don't need to reread this file again
        self._already_initialized = True

    def _raw_buffers(self, filters={}):
        # Create a dummy instance of DirectoryEntry in order to find out what
        # attributes an instance of DirectoryEntry has.
        dummy_dir_entry = DirectoryEntry(*[None for i in range(10)])

        # Remove invalid filter criteria
        for attr_name in filters.keys():
            if not hasattr(dummy_dir_entry, attr_name):
                warnings.warn('Removing not-existent DiretoryEntry '
                                'attribute "{:s}" from filter '
                                'criteria'.format(attr_name))
                del filters[attr_name]

        cursor = 0
        search_time = 0
        prev_cursor = 0
        dir_base = cursor
        dir_entries = []

        # The filters argument is a whitelist, so if it is non-empty, the
        # default action is to reject buffers.
        if filters:
            drop_buffer = True
        else:
            drop_buffer = False
        reached_end = False
        has_filename = False
        first_tag = True

        pbar = tqdm(desc='Reading File',
                    total=len(self._get_contents()),
                    unit='bytes')

        # This while loop is effectively *THE* inner loop of this module. It is
        # not unusual for this loop to run millions of times, so avoid doing
        # anything unnecessary within it. If possible, do any checks,
        # comparisons, unchanging calculations, etc. before entering.
        while not reached_end:
            try:
                dir_entry = DirectoryEntry.new_from_raw_data(
                    self._get_contents(), cursor)
            except IncompleteDirectoryEntryException:
                raise UnexpectedEndOfFileException(
                        'Reached EOF while reading DirectoryEntry starting at '
                        'byte {cursor:d} ({cursor:#x}) of file {filename:s}'
                        ''.format(cursor=cursor, filename=self.filename))

            # Save the directory entries so that they can be returned with the
            # buffer, eliminating the need to read the directories again when
            # interpreting the data.
            dir_entries.append(dir_entry)
            if first_tag:
                if ((dir_entry.tag_number == TIME_TAG) &
                        (dir_entry.number_of_bytes == 36) &
                        (dir_entry.samples == 2) &
                        (dir_entry.bytes_per_sample == 18)):
                    first_tag = False
                else:
                    search_time += 1
                    cursor += 1
                    dir_base = cursor
                    dir_entries = []
            else:


                # This if statement contains clauses for all of the reserved tag
                # number specified in the unser manuals for the M200 and M300,
                # although not all of them are relevant in this context.
                if dir_entry.tag_number == TIME_TAG:
                    pass

                elif dir_entry.tag_number == NEXT_TAG:
                    cursor = dir_base + dir_entry.data_offset

                    if not drop_buffer:
                        yield dir_entries, self._get_contents()[dir_base:cursor]

                    dir_base = cursor
                    dir_entries = []
                    first_tag = True

                    # The filters argument is a whitelist, so if it is non-empty, the
                    # default action is to reject buffers.
                    if filters:
                        drop_buffer = True
                    else:
                        drop_buffer = False

                elif RESERVED_TAG_LOW_END <= dir_entry.tag_number <= \
                        RESERVED_TAG_HIGH_END:
                    # reserved for future use
                    pass

                elif dir_entry.tag_number == FILENAME_TAG:
                    pass

                elif dir_entry.tag_number == FILEDATA_TAG:
                    pass

                elif (dir_entry.tag_number == LAST_TAG):
                    yield dir_entries, self._get_contents()[dir_base:]
                    reached_end = True


                if filters:
                    for attr in filters.keys():
                        # No check error checking is done on getattr() here because
                        # we already verified that all filter criteria correspond
                        # to DirectoryEntry attributes at the beginning of this
                        # method.
                        #
                        # The pre-validation is done to avoid unnecesary if
                        # statements or try/catch blocks in this loop.
                        if getattr(dir_entry, attr) in filters[attr]:
                            drop_buffer = False
                            break

                # This is a separate 'if' statement from the 'elif' above
                if dir_entry.tag_number != NEXT_TAG:
                    cursor += DirectoryEntry.SIZE

            if cursor >= len(self._get_contents()) - dir_entry.SIZE:
                yield dir_entries, self._get_contents()[dir_base:]
                reached_end = True

            pbar.update(cursor - prev_cursor)
            prev_cursor = cursor

        pbar.close()
        print(search_time)

    def _parse_acqtbl(self, acqtbl):
        """Parses an acquisition table and saves data about about tags to expect
        in the rest of the file

        Arguments:
            acqtbl -- A ConfigFile containing the entire contents of acquisition
                table

        Returns:
            n/a

        """
        # remove lines that are comments
        data = acqtbl.data.decode('utf-8')
        lines = [l.replace('\t', ' ').rstrip() for l in data.split('\n')
                    if l[0] != ';']

        reader = csv.reader(lines, dialect=self._ACQ_CSV_DIALECT)
        tags = [self._TAG_CLASS(*row) for row in reader]

        # TODO reimplement buffer types

        for tag in tags:
            self._add_tag(tag)

    def _parse_acq2tbl(self, acq2tbl):
        """Parses secondary acquisition table

        Arguments:
            acq2tbl -- A ConfigFile containing the entire contents of
                a secondary acquisition table

        Returns:
            n/a

        Implemetation Details:
            As of the writing of this documentation, this method only secondary
            tag numbers so that other methods know which tags and buffers to
            ignore.

        """
        comment_start = re.compile("^;")
        lines = [l for l in acq2tbl.data.split('\n')
                    if not comment_start.match(l)]

        secondary_tags = [int(l.split()[1]) for l in lines]

        self._secondary_tags.update(secondary_tags)

    def _unpack_string_data(self, raw_buffer, dir_entry):
        """Unpacks strings from a buffer.

        Unpacks a number of consecutive strings of a given length from a buffer
        of raw data.

        Arguments:
            raw_data -- a buffer of data from an M200 data file which contains
                the strings that need to be read
            dir_entry -- a DirectoryEntry with metadata pertaining to the string
                data

        Returns:
            A tuple of the strings that were read from the buffer

        """
        format_string = '<{:d}s'.format(dir_entry.bytes_per_sample) * \
            dir_entry.samples
        return struct.unpack_from(
            format_string, raw_buffer, dir_entry.data_offset)

    def iter_buffers(self, filters={}):
        """Iterates over each buffer in self.

        Arguments:
            n/a

        Returns:
            A generator that will iterate over each buffer in self.

        Warns:
            SEAReadWarning -- Warns if something unexpected/unsupported is
            found while reading through self.

        Generator:
            Yields:
                Each buffer of self, in order.

            Raises:
                StopIteration -- It's what iterators and generators do.

        """
        for dir_entries, raw_buffer in self._raw_buffers(filters):
            new_buffer = Buffer()
            if filters:
                discard_buffer = True
            else:
                discard_buffer = False

            for dir_entry in dir_entries:
                unknown_type = False

                if dir_entry.tag_number == TIME_TAG:
                    data = SEATime._make(
                        SEA_TIME_STRUCT.unpack_from(
                            raw_buffer, dir_entry.data_offset))

                    new_buffer.add_dataset(dir_entry, data)

                elif dir_entry.tag_number == NEXT_TAG:
                    if not discard_buffer:
                        # TODO reimplement buffer types
                        # # find which buffer type has the same set of tags as the
                        # # new buffer
                        # new_buffer_counter = collections.Counter(
                        #   [tn for tn in new_buffer.get_tag_numbers()
                        #       if tn not in [TIME_TAG, NEXT_TAG, LAST_TAG]
                        #       ])

                        # for buffer_type in self._buffer_types:
                        #   if new_buffer_counter == buffer_type.counter:
                        #       new_buffer_type = buffer_type
                        #       break
                        # else:
                        #   new_buffer_type = -1

                        # new_buffer.set_type(new_buffer_type)

                        # This 'yield' function makes this method a generator
                        # (google it). It allows us to only have one buffer
                        # object in memory at a time (versus creating a list of
                        # every buffer object and returning it at the end.)
                        yield new_buffer

                    break

                elif RESERVED_TAG_LOW_END <= dir_entry.tag_number <= \
                        RESERVED_TAG_HIGH_END:
                    # reserved for future use
                    warnings.warn('Found reserved tag '
                                    'number {:d}'.format(dir_entry.tag_number),
                                    SEAReadWarning)

                elif dir_entry.tag_number == FILENAME_TAG:
                    new_buffer.add_dataset(dir_entry,
                            self._unpack_string_data(raw_buffer, dir_entry))

                elif dir_entry.tag_number == FILEDATA_TAG:
                    new_buffer.add_dataset(dir_entry,
                            self._unpack_string_data(raw_buffer, dir_entry))

                elif dir_entry.tag_number == COMMAND_TAG:
                    discard_buffer = True
                elif dir_entry.tag_number == ERROR_TAG:
                    discard_buffer = True
                elif dir_entry.tag_number == SAME_TAG:
                    discard_buffer = True

                elif dir_entry.tag_number == LAST_TAG:
                    # We no longer need that reference to the binary file data
                    self._file_contents = None
                    # Reset the filters parameter since using mutable objects
                    # as default parameters can have some unexpected side
                    # effects
                    filters={}
                    # There are no more buffers to iterate over
                    return

                elif dir_entry.tag_number in self._secondary_tags:
                    # I haven't decided how to handle the secondary
                    # acquisitions, yet.
                    discard_buffer = True
                elif filters:
                    for attr in filters.keys():
                        if getattr(dir_entry, attr) in filters[attr]:
                            dataset = self._read_dataset(dir_entry, raw_buffer)
                            new_buffer.add_dataset(dir_entry, dataset)
                            discard_buffer = False
                else:
                    dataset = self._read_dataset(dir_entry, raw_buffer)

                    new_buffer.add_dataset(dir_entry, dataset)


    def _read_dataset_raw(self, dir_entry, raw_buffer, dataset_string):
        try:
            # The '<' at the start of the format string tells python
            # to read the data in little endian byte order.
            dataset = struct.unpack_from(
                '<{:s}'.format(dataset_string), raw_buffer,
                    dir_entry.data_offset)
        except Exception as e:
            print("Tag Number: {:d}\nAcq type: {:d}\n"
                    "Data Offset: {:d}\n# Samples: {:d}\n"
                    "B/Sample: {:d}\nStruct Size: {:d}\n"
                    "format: {:s}\n".format(
                        dir_entry.tag_number, dir_entry.typ,
                        dir_entry.data_offset, dir_entry.samples,
                        dir_entry.bytes_per_sample,
                        struct.calcsize(dataset_string),
                        's'))
            raise e

        return dataset

    def _read_dataset(self, dir_entry, raw_buffer):
        # Determine what the struct format string for this dataset
        # is. Most of the comments below are borrowed from the SEA
        # Model 200 User's Manual (ie m200.pdf).
        # Model 300 User's Manual (ie M300Reference.pdf).
        # ref: http://docs.python.org/2/library/struct.html
        # ref: m200.pdf
        # ref: M300Reference.pdf
        #
        # b - 8-bit signed integer
        # B - 8-bit unsigned integer
        # h - 16-bit signed integer
        # H - 16-bit unsigned integer
        # i - 32-bit signed integer
        # I - 32-bit unsigned integer
        # Q - 64-bit unsigned integer
        # Xs - a string of length X

        unknown_type = False

        if dir_entry.typ == 1:
            # Acquisition Type 1 (CAMAC Analog E205/E210)
            # 16-bit, two's complement integer
            sample_string = 'h'

        elif dir_entry.typ == 2:
            # Aquisition Type 2 (CAMAC 1D Counts)
            # 42 bytes representing 15 16-bit numbers followed by 2
            # 8-bit numbers. Since the 16-bit values are counts,
            # they should be unsigned.
            sample_string = '{:s}bb'.format('H' * 20)

        elif dir_entry.typ == 5:
            # Acquisition Type 5 (2D Mono Image)
            # This routine acquires the 4096 byte image block of the
            # 2D Mono probe. Each image contains 1024 32 bit slices.
            # Allocate 4096 bytes for this acquisition.
            sample_string = 'I' * 1024

        elif dir_entry.typ == 6:
            # Acquisition Type 6 (2D Mono TAS Factors)
            # This routine acquires two 16 bit words.
            sample_string = 'HH'

        elif dir_entry.typ == 7:
            # Acquisition Type 7 (2D Mono Elapsed Time)
            # This routine acquires a 32 bit word.
            sample_string = 'I'

        elif dir_entry.typ == 8:
            # Acquisition Type 8 (2D Mono Elapsed TAS/100)
            # This routine acquires a 32 bit word.
            sample_string = 'I'

        elif dir_entry.typ == 9:
            # Acquisition Type 9 (2D Mono Elapsed Shadow OR)
            sample_string = 'H'

        elif dir_entry.typ == 10:
            # Acquisiiton Type 10 (2D Mono Total Shadow OR)
            # This routine acquires a 16-bit, unsigned word.
            sample_string = 'H'

        elif dir_entry.typ == 11:
            # Acquisition Type 11 (2D Mono House Data)
            # This routine acquires 8 16-bit words.
            sample_string = 'H' * 8

        elif dir_entry.typ == 35:
            # Acquisition Type 35 (SEA Analog to Digital Input)
            # This rotine acquires two 8-bit bytes. The data acquired is in
            # two's complement integer coding.
            sample_string = 'h'

        elif dir_entry.typ == 66:
            # Acquisition Type 66 (2D Grey Advanced)
            # The final data size will not be larger than the number
            # of samples times the bytes per sample.
            # Each sample should be a set of 128-bit slices
            if dir_entry.bytes_per_sample % 16 != 0:
                warnings.warn(
                        '2D Grey Advanced sample size ({:d} bytes) '
                        'is not a multiple of 128 bits. The '
                        'resulting slices may me incorrect, '
                        'incomplete, misaligned, or otherwise '
                        'corrupt.'.format(
                            dir_entry.bytes_per_sample))

            num_slices = dir_entry.bytes_per_sample / 16
            # TODO The first 64-bit number will be
            try:
                sample_string = 'QQ' * num_slices
            except TypeError:
                sample_string = 'QQ' * math.ceil(num_slices)
                print(num_slices)

        elif dir_entry.typ == 78:
            # Acquisition Type 78 (CIP Image Data)
            # This data is stored in a compressed format, so just
            # read it as a bunch of bytes.
            sample_string = '{}H'.format('B' * 4096)

        elif dir_entry.typ == 85:
            # Acquisition Type 85 (Network Binary Data)
            # This is not necessarily stored in any particular format, so read
            # it as a bunch of bytes and let the receviver decode the data.
            sample_string = 'B' * dir_entry.bytes_per_sample

        elif dir_entry.typ == 255:
            # Acquisition Type 255 (Dummy Acquisition)
            sample_string = 'B' * dir_entry.bytes_per_sample

        else:
            # Unknown acquisition type, read it in as bytes
            warnings.warn(
                'Unknown acquisition type ({acq_type:d}) will be '
                'read as list of bytes. If you need to read data '
                'with acquisition type {acq_type:d}, you should '
                'update this program to recognize '
                'it.'.format(acq_type=dir_entry.typ),
                SEAReadWarning)

            dataset_string = 'B' * dir_entry.number_of_bytes
            unknown_type = True

        if not unknown_type:
            expected_size = dir_entry.samples * \
                    dir_entry.bytes_per_sample

            if dir_entry.number_of_bytes < expected_size:
                warnings.warn(
                    'Number of bytes in buffer ({:d}) is less than '
                    'expected ({:d}). This is legal in the file '
                    'format, but is not currently supported by this '
                    'program.'.format(
                        dir_entry.number_of_bytes, expected_size),
                    SEAReadWarning)
            elif dir_entry.data_offset + expected_size > len(raw_buffer):
                # If a dataset apparently goes past the end of the
                # buffer, raise an error.
                print(dir_entry.number_of_bytes, dir_entry.typ, dir_entry.samples, dir_entry.bytes_per_sample, expected_size, dir_entry.data_offset)
                warnings.warn(
                        'Expected end of dataset (0x{:x}) is beyond end '
                        'of buffer (0x{:x}).'.format(
                            dir_entry.data_offset + \
                                    expected_size, len(raw_buffer))
                        )
                raise ValueError
                return []


            dataset_string = sample_string * dir_entry.samples

        return self._read_dataset_raw(dir_entry, raw_buffer, dataset_string)

class BufferType(object):
    def __init__(self, number, tags):
        self.number = number
        self.tags = tuple(tags)
        self.counter = collections.Counter(
                                [t.tag_number for t in self.tags])


    def __hash__(self):
        return self.number

    def __cmp__(self, other):
        return self.number - other.number

    def make_buffer(self):
        buff = Buffer()

        buff.number = self.number
        buff.tags = self.tags

        return buff


class Buffer(object):
    """Software representaion of a single buffer in an M200 file

    Methods:
        set_type -- set the BufferType of this Buffer
        get_type -- get the BufferType of this Buffer
        add_dataset -- add a dataset from a buffer in an M200 file to this
            Buffer
        get_dir_entry_by_tag_number -- fetch a DirectoryEntry from this Buffer
        get_tag_numbers -- get a list of tag numbers for all DirectoryEntries
            associated with this Buffer
        get_dir_entries_by_typ -- get a list of all DirectoryEntries in this
            Buffer with a given typ(e)
        get_dataset_by_tag_number -- get the dataset from this Buffer that has a
            given tag number
        get_datasets_by_typ -- get a list of all datasets in this Buffer that
            have a given typ(e)

    """
    def __init__(self):
        # Since Python uses object references (ie won't consume
        # memory with a bajillion copies) and since the buffer will
        # be queried based on multiple criteria, we will use a number
        # of dicts in order to provide lookups with complexity O(1).
        self.dir_entries_by_tag_number = {}
        self.dir_entries_by_typ = {}
        self.datasets_by_tag_number = {}
        self.datasets_by_typ = {}

        self.buffer_type = None

    def set_type(self, buffer_type):
        self.buffer_type = buffer_type

    def get_type(self):
        return self.buffer_type

    def add_dataset(self, dir_entry, dataset):
        self.dir_entries_by_tag_number[dir_entry.tag_number] = dir_entry
        self.datasets_by_tag_number[dir_entry.tag_number] = dataset

        # In diretory entries for time data in asynchronous buffers, the
        # acquisition type (apparently) contains the acquisition type of the
        # buffer's asynchronous master event, rather than the acquisition type
        # of the time. To prevent this from interfering with retrieval of data
        # by type, we will not store time data in the dicts which are indexed by
        # acquisition type.
        #
        # Since there should only ever be a single time dataset in a buffer,
        # retrieval by tag number is still possible (and makes sense) for time
        # data, especially since time data will always have the same tag number.
        if dir_entry.tag_number != TIME_TAG:
            try:
                self.dir_entries_by_typ[dir_entry.typ].append(dir_entry)
            except KeyError as ke:
                if ke.args[0] == dir_entry.typ:
                    self.dir_entries_by_typ[dir_entry.typ] = [dir_entry]
                else:
                    raise ke

            try:
                self.datasets_by_typ[dir_entry.typ].append(dataset)
            except KeyError as ke:
                if ke.args[0] == dir_entry.typ:
                    self.datasets_by_typ[dir_entry.typ] = [dataset]
                else:
                    raise ke

    def get_dir_entry_by_tag_number(self, tag_number):
        return self.dir_entries_by_tag_number[tag_number]

    def get_tag_numbers(self):
        return self.dir_entries_by_tag_number.keys()

    def get_dir_entries_by_typ(self, typ):
        return self.dir_entries_by_typ[typ]

    def get_dataset_by_tag_number(self, tag_number):
        return self.datasets_by_tag_number[tag_number]

    def get_datasets_by_typ(self, typ):
        return self.datasets_by_typ[typ]

    def __str__(self):
        strings = []
        strings.append('Buffer (Type {})'.format(self.buffer_type.number
            if self.buffer_type is not None else 'unknown'))

        for dir_entry in self.dir_entries_by_tag_number.values():
            strings.append(str(dir_entry))

        return '\n\t'.join(strings)


class DirectoryEntry(object):
    struct = struct.Struct('<HHHHHBBBBH')
    SIZE = 16

    def __init__(self, tag_number, data_offset, number_of_bytes, samples,
            bytes_per_sample, typ, parameter1, parameter2, parameter3, address):
        self.tag_number = tag_number
        self.data_offset = data_offset
        self.number_of_bytes = number_of_bytes
        self.samples = samples
        self.bytes_per_sample = bytes_per_sample
        self.typ = typ
        self.parameter1 = parameter1
        self.parameter2 = parameter2
        self.parameter3 = parameter3
        self.address = address

    def __repr__(self):
        return ("<sea.DirectoryEntry tag_number={tag_number:d} "
            "data_offset={data_offset:d} number_of_bytes={number_of_bytes:d} "
            "samples={samples:d} bytes_per_sample={bytes_per_sample:d} "
            "type={typ:d} parameter1={parameter1:d} parameter2={parameter2:d} "
            "parameter3={parameter3:d} address={address:#x}"
            "".format(**vars(self)))

    @classmethod
    def new_from_raw_data(cls, file_contents, start=0):
        try:
            return cls(*cls.struct.unpack_from(file_contents, start))
        except struct.error as se:
            if str(se) == 'unpack_from requires a buffer of at least 16 bytes':
                raise IncompleteDirectoryEntryException
            else:
                raise se


class ConfigFile(object):
    def __init__(self, name=None, data=None):
        self.name = name
        self.data = data



"""Probe types module."""

from abc import ABC, abstractmethod
from enum import Enum


class ProbeType(Enum):
    """Enumerate for defining a probe type."""

    Unknown = 0
    TwoDs = 1
    TwoDc = 2
    Hvps = 3
    Cip = 4
    Pip = 5
    Cpi = 6
    Hsi = 7
    Hvps4Res50 = 8
    Hvps4Res150 = 9


class SpifProbe(ABC):
    """
    Probe type interface.

    Each probe must inherit this interface.

    Parameters
    ----------
    ABC : abc.ABCMeta
        Metaclass to implement abstract classes in python
    """

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, 'CPI', '2DS-V', 'CIP', etc.
        """
        pass

    @classmethod
    def get_name_lower(cls) -> str:
        """
        Get a probe type name in lower case.

        Returns
        -------
        str
            Probe type in lower case, 'cpi', '2ds-v', 'cip', etc.
        """
        return cls.get_name().lower()

    @classmethod
    @abstractmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        pass

    @classmethod
    @abstractmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Unknown


class CpiProbe(SpifProbe):
    """Cpi probe class."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, 'CPI'
        """
        return "CPI"

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 2.3

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Cpi


class TwoDsProbe(SpifProbe):
    """2DS probe class."""

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 10.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.TwoDs


class TwoDsVProbe(TwoDsProbe):
    """2DS probe class. Channel V."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, '2DS-V'
        """
        return "2DS-V"


class TwoDsHProbe(TwoDsProbe):
    """2DS probe class. Channel H."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, '2DS-H'
        """
        return "2DS-H"


class HvpsProbe(SpifProbe):
    """HVPS probe class."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `HVPS`
        """
        return "HVPS"

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 150.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Hvps


class Hvps4Res50(SpifProbe):
    """HVPS-4 probe class, resolution 50 micron."""

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 50.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Hvps4Res50


class Hvps4Res150(SpifProbe):
    """HVPS-4 probe class, resolution 150 micron."""

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 150.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Hvps4Res150


class Hvps4V50Probe(Hvps4Res50):
    """HVPS-4 probe class. Channel V, resolution 50"""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `HVPS4-V50`
        """
        return "HVPS4-V50"


class Hvps4H50Probe(Hvps4Res50):
    """HVPS-4 probe class. Channel H, resolution 50"""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `HVPS4-H50`
        """
        return "HVPS4-H50"


class Hvps4V150Probe(Hvps4Res150):
    """HVPS-4 probe class. Channel V, resolution 150"""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `HVPS4-V150`
        """
        return "HVPS4-V150"


class Hvps4H150Probe(Hvps4Res150):
    """HVPS-4 probe class. Channel H, resolution 150"""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `HVPS4-H150`
        """
        return "HVPS4-H150"


class CipProbe(SpifProbe):
    """CIP probe class."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `CIP`
        """
        return "CIP"

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 25.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Cip


class PipProbe(SpifProbe):
    """PIP probe class."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `PIP`
        """
        return "PIP"

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 100.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Pip


class TwoDcProbe(SpifProbe):
    """2DC probe class."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, `2DC`
        """
        return "2DC"

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 25.0

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.TwoDc


class HsiProbe(SpifProbe):
    """Hsi probe class."""

    @classmethod
    def get_name(cls):
        """
        Get a probe type name.

        Returns
        -------
        str
            Probe type, 'HSI'
        """
        return "HSI"

    @classmethod
    def get_resolution(cls) -> float:
        """
        Get a probe resolution.

        Returns
        -------
        float
            Probe resolution, microns per pixel
        """
        return 2.5

    @classmethod
    def get_type(cls) -> ProbeType:
        """
        Get a probe type.

        Returns
        -------
        ProbeType
            Probe type
        """
        return ProbeType.Hsi



"""Factory method to get a probe instance based on the probe type name."""

from .probetypes import (
    CipProbe,
    CpiProbe,
    HsiProbe,
    HvpsProbe,
    Hvps4V50Probe,
    Hvps4V150Probe,
    Hvps4H50Probe,
    Hvps4H150Probe,
    PipProbe,
    SpifProbe,
    TwoDcProbe,
    TwoDsHProbe,
    TwoDsVProbe,
)


def get_probe(probe_type: str) -> SpifProbe:
    """
    Get a probe instance from a probe type name.

    Parameters
    ----------
    probe_type : str
        Probe type name

    Returns
    -------
    SpifProbe
        Probe instance

    Raises
    ------
    ValueError
        Invalid probe type
    """
    probe = probe_type.lower()

    if probe == "cpi":
        return CpiProbe()
    elif probe == "hsi":
        return HsiProbe()
    elif probe == "2ds-v":
        return TwoDsVProbe()
    elif probe == "2ds-h":
        return TwoDsHProbe()
    elif probe == "2dc":
        return TwoDcProbe()
    elif probe == "hvps":
        return HvpsProbe()
    elif probe == "hvps4-v50":
        return Hvps4V50Probe()
    elif probe == "hvps4-h50":
        return Hvps4H50Probe()
    elif probe == "hvps4-v150":
        return Hvps4V150Probe()
    elif probe == "hvps4-h150":
        return Hvps4H150Probe()
    elif probe == "cip":
        return CipProbe()
    elif probe == "pip":
        return PipProbe()
    else:
        raise ValueError("Invalid probe type.")



import configparser
from datetime import datetime
import pathlib as pl

from copy import deepcopy

import click

import numpy as np

import netCDF4 as nc
import xarray as xr

from spifpy.spif import SPIFFile


@click.command()
@click.argument(
    "filename",
    type=click.Path(
        exists=True,  # File needs to exist
        file_okay=True,  # Can provide a file as argument
        dir_okay=False,  # Cannot provide a directory as an argument
        path_type=pl.Path,  # Pre-emptively convert to a pl.Path object
    ),
)
@click.argument(
    "config_file",
    type=click.Path(
        exists=True,  # File needs to exist
        file_okay=True,  # Can provide a file as argument
        dir_okay=False,  # Cannot provide a directory as an argument
        path_type=pl.Path,  # Pre-emptively convert to a pl.Path object
    ),
)
@click.argument("start", metavar='START', type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S"]))
@click.argument("end", metavar='END', type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S"]))
def cut(filename, config_file, start, end):
    """
        Cut a spif file from start-time to end-time.
        Both times are in format of %Y-%m-%dT%H:%M:%S
    """

    # Args checker

    # Args transformer

    # Processing
    file_cutter = SPIFCutDatasetGenerator(filename, config_file, start, end)

    file_cutter.slice_dataset()

    file_cutter.save()

class SPIFCutDatasetGenerator:
    def __init__(self, filename, config_file, start, end) -> None:

        self.filename = filename
        self.config_file = config_file
        self.start = start
        self.end = end

        self.cut_file = self.create_cut_file()
        self.add_variables_to_cut_file_groups()

    @property
    def cut_filename(self):
        start_str = self.start.strftime("%H%M%S")
        end_str = self.end.strftime("%H%M%S")

        return f"{self.filename.resolve().parent}/{self.config_file.stem}_{start_str}_{end_str}.nc"

    @property
    def file_start_date(self):
        data = nc.Dataset(self.filename, mode="r")
        start_date = getattr(data,"start_date").strip()
        data.close()

        return datetime.strptime(start_date, "%Y-%m-%d %H:%M:%S")

    @property
    def file_groups(self):
        data = nc.Dataset(self.filename, mode="r")
        groups = [k for k in data.groups.keys()]
        data.close()

        return groups

    @property
    def file_config(self):

        config = configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str

        config.read(self.config_file)

        return config

    def create_cut_file(self):
        cut_file = SPIFFile( self.cut_filename, self.file_config )
        cut_file.create_file()

        for g in self.file_groups:
            cut_file.create_inst_group(g)

        cut_file.set_start_date(self.file_start_date.__str__())

        return cut_file

    def add_variables_to_cut_file_groups(self):
        for g in self.file_groups:
            original_data = xr.open_dataset( self.filename, group = f"{g}/core", decode_cf = False )

            for v in original_data:
                dataset = self.cut_file.rootgrp.groups[g].groups['core']

                if v not in dataset.variables :
                    self.cut_file.create_variable(dataset, v, original_data[v].dtype, original_data[v].dims, attrs=original_data[v].attrs)

            original_data.close()

    def slice_dataset(self):
        for g in self.file_groups:
            cut_group = self.slice_group(g)

            core_grp = self.cut_file.rootgrp.groups[g].groups['core']

            for v in cut_group:
                self.cut_file.write_variable(core_grp, v, cut_group[v].data)

    def slice_group(self, group):
        data = xr.open_dataset(self.filename, group = f"{group}/core", decode_cf = False)

        # datetime to seconds since start date
        imgsec_start = int((self.start-self.file_start_date).total_seconds())
        imgsec_end = int((self.end-self.file_start_date).total_seconds())

        # image start and end indices
        image_start = np.searchsorted(data['image_sec'].values, imgsec_start, side='left')-1
        image_end = np.searchsorted(data['image_sec'].values, imgsec_end, side='right')

        # pixel start and end indices
        pixel_start = int(np.sum(data['image_len'].values[0:image_start])*128)
        pixel_end = int(np.sum(data['image_len'].values[0:image_end-1])*128)

        # buffer start and end indices
        buffer_start = data['buffer_index'].values[image_start]
        buffer_end = data['buffer_index'].values[image_end]

        # cut data
        data_cut = \
            data.isel( Pixels  = slice( pixel_start,  pixel_end  ) ) \
                .isel( Images  = slice( image_start,  image_end  ) ) \
                .isel( Buffers = slice( buffer_start, buffer_end ) )

        data_cut.attrs = deepcopy(data.attrs)

        data.close()

        return data_cut

    def save(self):
        self.cut_file.close()



import numpy as np
from numba import jit

from spifpy.spif import TIME_CHUNK

# The size of the metadata in a particle record
# Word 1 = Flag 2S
# Word 2 = word for h image metadata
# Word 3 = word for v image metadata
# Word 4 = word for particle count
# Word 5 = word for number of slices

METADATA_LENGTH = 5

# Offsets to find specific metadata in an image record

WORD_H_OFFSET = 1
WORD_V_OFFSET = 2
WORD_PC_OFFSET = 3
WORD_NUM_SLICE_OFFSET = 4

# Easier to define here than to have to flip 1 and 0 in the code

SHADED_VAL = 0
CLEAR_VAL = 1

# Useful datatypes

decoded_word_type = np.dtype([
            ("is_image_slice", "u2"),
            ("is_start_slice", "u2"),
            ("num_shaded", "u2"),
            ("num_clear", "u2")
        ])

class ImageMetadataContainer:

    def __init__(self):
        self.buffer_idx = 0

        self.n_h = 0
        self.timing_h = 0
        self.mismatch_h = 0
        self.fifo_h = 0
        self.overload_h = 0

        self.n_v = 0
        self.timing_v = 0
        self.mismatch_v = 0
        self.fifo_v = 0
        self.overload_v = 0

        self.particle_count = 0
        self.num_slices = 0

        self.h_start = 0
        self.h_end = 0

        self.v_start = 0
        self.v_end = 0

        self.frame_len = 0
        self.image_in_buffer = 0

class ImageMetadataProcessor:
    """

    This is for words 2 in a particle frame

    NH (Word 2)
    -----------------------------------------------------------

    Bits 0–11 Number of horizontal words–Includes Timing Words if present
    Bit 12 – 1 = Timing Words not found
    Bit 13 – Timing Word mismatch
    Bit 14 — FIFO Empty (means the next particle was cut off)
    Bit 15 – The last two words of the horizontal data record are overload timing words

    NV (Word 3)
    -------------------------------------------------------------

    Bits 0 –11 Number of vertical words–Includes Timing Words if not same as  the  horizontal Timing Word and the TW were found.
    Bit 12 –1 = Timing Words not found
    Bit 13 –Timing Word mismatch
    Bit 14-FIFO Empty before timing word found
    Bit 15 –The last two words of the vertical data record are overload timing words

    """

    def __init__(self) -> None:
        pass

    def process_metadata(self, buffer_idx, buffer):

        metadata = ImageMetadataContainer()

        metadata.buffer_idx = buffer_idx

        metadata.n_h = num_words(buffer[buffer_idx + WORD_H_OFFSET])
        metadata.timing_h = timing_words_not_found(buffer[buffer_idx + WORD_H_OFFSET])
        metadata.mismatch_h = timing_word_mismatch(buffer[buffer_idx + WORD_H_OFFSET])
        metadata.fifo_h = fifo_empty(buffer[buffer_idx + WORD_H_OFFSET])
        metadata.overload_h = overload_timing_words_exist(buffer[buffer_idx + WORD_H_OFFSET])

        metadata.n_v = num_words(buffer[buffer_idx + WORD_V_OFFSET])
        metadata.timing_v = timing_words_not_found(buffer[buffer_idx + WORD_V_OFFSET])
        metadata.mismatch_v = timing_word_mismatch(buffer[buffer_idx + WORD_V_OFFSET])
        metadata.fifo_v = fifo_empty(buffer[buffer_idx + WORD_V_OFFSET])
        metadata.overload_v = overload_timing_words_exist(buffer[buffer_idx + WORD_V_OFFSET])

        metadata.particle_count = buffer[buffer_idx + WORD_PC_OFFSET]
        metadata.num_slices = buffer[buffer_idx + WORD_NUM_SLICE_OFFSET]

        metadata.h_start = metadata.buffer_idx + METADATA_LENGTH
        metadata.h_end = metadata.h_start + metadata.n_h

        metadata.v_start = metadata.buffer_idx + METADATA_LENGTH + metadata.n_h
        metadata.v_end = metadata.v_start + metadata.n_v

        metadata.frame_len = METADATA_LENGTH + metadata.n_h + metadata.n_v
        metadata.image_in_buffer = (metadata.buffer_idx + metadata.frame_len) < 2048

        return metadata

@jit(nopython = True)
def num_words(word):
    # Bit masking out of a 16-bit number
    # to only get the 12 bit component
    return word & 0b0000111111111111

@jit(nopython = True)
def timing_words_not_found(word):
    # Bitmask to get 12th bit only, then bit shift right
    # 12 spots to keep only that bit
    return (word & 0b0001000000000000) >> 12

@jit(nopython = True)
def timing_word_mismatch(word):
    # Bitmask to get 13th bit only, then bit shift right
    # 13 spots to keep only that bit
    return (word & 0b0010000000000000) >> 13

@jit(nopython = True)
def fifo_empty(word):
    # Bitmask to get 14th bit only, then bit shift right
    # 14 spots to keep only that bit
    return (word & 0b0100000000000000) >> 14

@jit(nopython = True)
def overload_timing_words_exist(word):
    # Bitmask to get 15th bit only, then bit shift right
    # 15 spots to keep only that bit
    return (word & 0b1000000000000000) >> 15

class RawImageContainer:

    def __init__(self) -> None:
        self.raw_image_h = np.array([], dtype=np.uint16)
        self.raw_image_v = np.array([], dtype=np.uint16)

class ImageTimewordContainer:

    def __init__(self) -> None:
        self.timeword_h_upper = 0
        self.timeword_h_lower = 0
        self.timeword_v_upper = 0
        self.timeword_v_lower = 0

class RawImageExtractor:

    def __init__(self) -> None:
        self.raw_image_container = RawImageContainer()
        self.image_timeword_container = ImageTimewordContainer()

    def extract_raw_images(self, metadata, buffer):
        self.raw_image_container = RawImageContainer()

        raw_image_h = buffer[metadata.h_start:metadata.h_end]
        raw_image_v = buffer[metadata.v_start:metadata.v_end]

        if metadata.timing_h == 0:
            raw_image_h = raw_image_h[:-2]

        if metadata.timing_v == 0:
            raw_image_v = raw_image_v[:-2]

        self.raw_image_container.raw_image_h = raw_image_h
        self.raw_image_container.raw_image_v = raw_image_v

        return self.raw_image_container

    def extract_image_timewords(self, metadata, buffer):
        self.image_timeword_container = ImageTimewordContainer()

        raw_image_h = buffer[metadata.h_start:metadata.h_end]
        raw_image_v = buffer[metadata.v_start:metadata.v_end]

        if (metadata.timing_h == 0) and (len(raw_image_h) >= 2):
            self.image_timeword_container.timeword_h_upper = raw_image_h[-2]
            self.image_timeword_container.timeword_h_lower = raw_image_h[-1]

        if (metadata.timing_v == 0) and (len(raw_image_v) >= 2):
            self.image_timeword_container.timeword_v_upper = raw_image_v[-2]
            self.image_timeword_container.timeword_v_lower = raw_image_v[-1]

        return self.image_timeword_container

class DecodedImageContainer:

    def __init__(self) -> None:
        self.decoded_image_h = np.empty(0, dtype = decoded_word_type)
        self.decoded_image_v = np.empty(0, dtype = decoded_word_type)

class RawImageDecoder:

    def __init__(self) -> None:
        pass

    def decode_dual_channel_images(self, raw_image_container):

        decoded_image_container = DecodedImageContainer()
        decoded_image_container.decoded_image_h = decode_image(raw_image_container.raw_image_h)
        decoded_image_container.decoded_image_v = decode_image(raw_image_container.raw_image_v)

        return decoded_image_container

@jit
def decode_image(encoded_image):
    decoded_image = np.zeros(len(encoded_image), dtype = decoded_word_type)

    for i, word in enumerate(encoded_image):
        if word == 0x7fff:
            decoded_image['is_image_slice'][i] = 1
            decoded_image['is_start_slice'][i] = 1
            decoded_image['num_clear'][i] = 128
            decoded_image['num_shaded'][i] = 0
        elif word == 0x4000:
            decoded_image['is_image_slice'][i] = 1
            decoded_image['is_start_slice'][i] = 1
            decoded_image['num_clear'][i] = 0
            decoded_image['num_shaded'][i] = 128
        else:
            decoded_image['is_image_slice'][i] = ((word & 2**15) >> 15) == 0
            decoded_image['is_start_slice'][i] = (word & 2**14) >> 14
            decoded_image['num_shaded'][i] = (word & 0b0011111110000000) >> 7
            decoded_image['num_clear'][i]  = (word & 0b0000000001111111)

    valid_image_words = decoded_image['is_image_slice'] == True

    return decoded_image[valid_image_words]

class DecompressedImageContainer:

    def __init__(self) -> None:
        self.decompressed_image_h = np.array([], np.uint8)
        self.decompressed_image_v = np.array([], np.uint8)

class DecodedImageDecompressor:

    def __init__(self) -> None:
        pass

    def decompress_image(self, decoded_image_container):

        decompressed_image_container = DecompressedImageContainer()

        decompressed_image_container.decompressed_image_h = self.decompress_single_channel_image(decoded_image_container.decoded_image_h)
        decompressed_image_container.decompressed_image_v = self.decompress_single_channel_image(decoded_image_container.decoded_image_v)

        return decompressed_image_container

    def decompress_single_channel_image(self, decoded_image):

        if len(decoded_image) == 0:
            return []
        else:
            return decompress_complete_image(decoded_image)

@jit(nopython = True)
def get_complete_image_slice_inds(start_slice_flags):
    image_slice_id = np.cumsum(start_slice_flags)

    image_slice_inds = []

    for i in np.unique(image_slice_id):
        image_slice_inds.append(
            np.ravel(
                np.argwhere(image_slice_id == i)
            )
        )

    return image_slice_inds

@jit(nopython = True)
def decompress_complete_image(decoded_image):
    image_slice_inds = get_complete_image_slice_inds(decoded_image['is_start_slice'])

    image_slices = [int(x) for x in range(0)]
    #breakpoint()
    for i, slice_collection in enumerate(image_slice_inds):
        image_slice = [int(x) for x in range(0)]

        for slice_idx in slice_collection:

            """
            Be sure to append clear, THEN shaded
            Order matters very much here
            """

            image_slice += [CLEAR_VAL]*decoded_image['num_clear'][slice_idx]
            image_slice += [SHADED_VAL]*decoded_image['num_shaded'][slice_idx]

        # Add some clear bits to any incomplete slices

        if len(image_slice) < 128:
            image_slice += [CLEAR_VAL] * (128 - len(image_slice))

        # NOTE -> This is a TEMPORARY fix to a very strange encountered problem. To be sorted out later.

        if len(image_slice) > 128:
            image_slice += [CLEAR_VAL] * int(128*np.ceil(len(image_slice)/128) - len(image_slice) )

        image_slices += image_slice

    return np.array(image_slices, dtype=np.uint8)

class AssembledImageRecordContainer:

    def __init__(self) -> None:

        # Critical variables for the SPIF format

        self.buffer_id = np.array([], np.int32)
        self.image_sec = np.array([], np.int32)
        self.image_ns = np.array([], np.int32)

        self.image_len = np.array([], np.int32)
        self.image = np.array([], dtype = np.uint8)

        # Auxiliaries not neccessary for the format

        self.num_words = np.array([], dtype = np.uint16)
        self.timing_flag = np.array([], dtype = np.uint16)
        self.mismatch_flag = np.array([], dtype = np.uint16)
        self.fifo_flag = np.array([], dtype = np.uint16)
        self.overload_flag = np.array([], dtype = np.uint16)

        self.particle_count = np.array([], dtype = np.uint16)
        self.num_slices = np.array([], dtype = np.uint16)

        self.timeword_upper = np.array([], dtype = np.uint16)
        self.timeword_lower = np.array([], dtype = np.uint16)

        self.tas = np.array([], dtype = np.float32)

class ImageRecordAssembler:

    def __init__(self) -> None:
        pass

    def assemble_images(
        self,
        buffer_id,
        buffer_sec,
        buffer_ns,
        metadata_containers,
        timeword_containers,
        decompressed_image_containers,
        housekeeping
    ):

        image_container_h = AssembledImageRecordContainer()
        image_container_v = AssembledImageRecordContainer()

        image_container_h, image_container_v = self.set_buffer_info(
            buffer_id = buffer_id,
            buffer_sec = buffer_sec,
            buffer_ns = buffer_ns,
            num_images = len(metadata_containers),
            image_container_h = image_container_h,
            image_container_v = image_container_v
        )

        image_container_h, image_container_v = self.set_image_tas(
            housekeeping = housekeeping,
            num_images = len(metadata_containers),
            image_container_h = image_container_h,
            image_container_v = image_container_v
        )

        if len(metadata_containers) > 0:

            image_container_h, image_container_v = self.set_images(
                image_container_h,
                image_container_v,
                decompressed_image_containers
            )

            image_container_h, image_container_v = self.set_auxiliaries(
                metadata_containers,
                timeword_containers,
                image_container_h,
                image_container_v
            )

        return image_container_h, image_container_v

    def set_buffer_info(self, buffer_id, buffer_sec, buffer_ns, num_images, image_container_h, image_container_v):

        image_container_h.buffer_id = np.zeros(num_images, dtype = np.int32) + buffer_id
        image_container_h.image_sec = np.zeros(num_images, dtype = np.int32) + buffer_sec
        image_container_h.image_ns = np.zeros(num_images, dtype = np.int32) + buffer_ns

        image_container_v.buffer_id = np.zeros(num_images, dtype = np.int32) + buffer_id
        image_container_v.image_sec = np.zeros(num_images, dtype = np.int32) + buffer_sec
        image_container_v.image_ns = np.zeros(num_images, dtype = np.int32) + buffer_ns

        return image_container_h, image_container_v

    def set_image_tas(self, housekeeping, num_images, image_container_h, image_container_v):

        housekeeping = housekeeping[ housekeeping['incomplete_packet'] == 0 ]

        if len(housekeeping) == 0:
            tas = np.zeros(num_images)*np.nan
        else:
            tas = np.zeros(num_images, dtype = np.float32) + housekeeping['tas'][-1]

        image_container_h.tas = tas
        image_container_v.tas = tas

        return image_container_h, image_container_v

    def set_images(
        self,
        image_container_h,
        image_container_v,
        decompressed_image_containers
    ):
        image_container_h.image_len = np.array(
            [len(x.decompressed_image_h) / 128 for x in decompressed_image_containers],
            dtype = np.int32
        )

        image_container_v.image_len = np.array(
            [len(x.decompressed_image_v) / 128 for x in decompressed_image_containers],
            dtype = np.int32
        )

        image_container_h.image = np.concatenate(
            [x.decompressed_image_h for x in decompressed_image_containers]
        )

        image_container_v.image = np.concatenate(
            [x.decompressed_image_v for x in decompressed_image_containers]
        )

        return image_container_h, image_container_v

    def set_auxiliaries(
        self,
        metadata,
        timewords,
        image_container_h,
        image_container_v
    ):

        image_container_h.num_words      = np.array([x.n_h for x in metadata], dtype = np.uint16)
        image_container_h.timing_flag    = np.array([x.timing_h for x in metadata], dtype = np.uint16)
        image_container_h.mismatch_flag  = np.array([x.mismatch_h for x in metadata], dtype = np.uint16)
        image_container_h.fifo_flag      = np.array([x.fifo_h for x in metadata], dtype = np.uint16)
        image_container_h.overload_flag  = np.array([x.overload_h for x in metadata], dtype = np.uint16)
        image_container_h.particle_count = np.array([x.particle_count for x in metadata], dtype = np.uint16)
        image_container_h.num_slices     = np.array([x.num_slices for x in metadata], dtype = np.uint16)
        image_container_h.timeword_upper = np.array([x.timeword_h_upper for x in timewords], dtype = np.uint16)
        image_container_h.timeword_lower = np.array([x.timeword_h_lower for x in timewords], dtype = np.uint16)

        image_container_v.num_words      = np.array([x.n_v for x in metadata], dtype = np.uint16)
        image_container_v.timing_flag    = np.array([x.timing_v for x in metadata], dtype = np.uint16)
        image_container_v.mismatch_flag  = np.array([x.mismatch_v for x in metadata], dtype = np.uint16)
        image_container_v.fifo_flag      = np.array([x.fifo_v for x in metadata], dtype = np.uint16)
        image_container_v.overload_flag  = np.array([x.overload_v for x in metadata], dtype = np.uint16)
        image_container_v.particle_count = np.array([x.particle_count for x in metadata], dtype = np.uint16)
        image_container_v.num_slices     = np.array([x.num_slices for x in metadata], dtype = np.uint16)
        image_container_v.timeword_upper = np.array([x.timeword_v_upper for x in timewords], dtype = np.uint16)
        image_container_v.timeword_lower = np.array([x.timeword_v_lower for x in timewords], dtype = np.uint16)

        return image_container_h, image_container_v

# A utility method to define the extra image related variables in a netcdf file

def add_auxiliary_core_variables(spiffile, inst_name):
        coregrp = spiffile.rootgrp[inst_name]['core']

        spiffile.create_variable(
            coregrp,
            'num_words',
            'u2',
            ('Images',),
            {
                'long_name':'Number of data words present in image',
                'units':'counts'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'timing_flag',
            'u2',
            ('Images',),
            {
                'long_name':'Timing flag for image',
                'units':'boolean'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'mismatch_flag',
            'u2',
            ('Images',),
            {
                'long_name':'Mismatch flag for image',
                'units':'boolean'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'fifo_flag',
            'u2',
            ('Images',),
            {
                'long_name':'FIFO flag for image',
                'units':'boolean'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'overload_flag',
            'u2',
            ('Images',),
            {
                'long_name':'Overload flag for image',
                'units':'binary'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'particle_count',
            'u2',
            ('Images',),
            {
                'long_name':'Number of particles detected in image',
                'units':'counts'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'num_slices',
            'u2',
            ('Images',),
            {
                'long_name':'Number of slices detected in the image',
                'units':'counts'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'timeword_upper',
            'u2',
            ('Images',),
            {
                'long_name':'Upper 16 bits of timeword for image',
                'units':'clock ticks'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'timeword_lower',
            'u2',
            ('Images',),
            {
                'long_name':'Lower 16 bits of timeword for image',
                'units':'clock ticks'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'tas',
            'f',
            ('Images',),
            {
                'long_name':'True airspeed as recorded by probe',
                'units':'m/s'
            },
            chunksizes=(TIME_CHUNK,)
        )



#!/usr/bin/env python
# coding: utf-8

import argparse
import configparser
import datetime
import glob
import os
import shutil
import sys
import pathlib as pl

import netCDF4 as nc

import numpy

def addaux():
    parser = get_parser()
    args = parser.parse_args()

    args_checker = ArgsChecker(args)
    args_checker.check_args()

    aux_checker = AuxFileChecker(args.aux_filename, args.config_filename)
    aux_checker.check_aux_file()

    args_transformer = ArgsTransformer(args)

    transformed_args = args_transformer.transform_args()

    add_aux_to_spif(
        transformed_args['spif_filename'],
        transformed_args['aux_filename'],
        transformed_args['inst_names'],
        transformed_args['config_filename'],
        transformed_args['output_filename']
    )

def get_parser():
    parser = argparse.ArgumentParser(description='Adds auxiliary data'
                                                 ' to SPIF file.')

    parser.add_argument('spif_filename',
                        type=str,
                        help='path to SPIF file to modify. ' +
                             'Accepts wildcard characters to process ' +
                             'directories if argument is enclosed in quotes.')

    parser.add_argument('aux_filename',
                        type=str,
                        help='path to source file for aux data')

    parser.add_argument('-i',
                        dest='inst_names',
                        type=str,
                        help='instrument group to apply aux data to -- optional.' +
                             ' If none provided, aux data will be added to all' +
                             ' instruments in SPIF file',
                        default='all')

    parser.add_argument('-c', '--config-filename',
                        dest='config_filename',
                        type=str,
                        help='path to config file for aux data',
                        default=None)

    parser.add_argument('-o',
                        dest='output_filename',
                        type=str,
                        help='when provided, spif file is copied to directory' +
                             ' and aux data is added to new file.',
                        default=None)

    return parser


class ArgsChecker:
    def __init__(self, args):
        self.args = args

    def check_args(self):
        self.check_filename(self.args.spif_filename)

        self.check_aux_filename(self.args.aux_filename)

        if self.args.inst_names != 'all':
            self.check_inst_name(self.args.spif_filename, self.args.inst_names)

        self.check_config_filename(self.args.config_filename)

        if self.args.output_filename is not None:
            self.check_output_filename(self.args.output_filename)

    def check_filename(self, filename):

        try:
            assert pl.Path(filename).is_file()
        except AssertionError:
            print(f"ERROR : The file {filename} is not a valid file")
            raise

        try:
            assert pl.Path(filename).suffix == '.nc'
        except AssertionError:
            print(f"ERROR : The file {filename} does not have a .nc filename ending")
            raise

        try:
            assert pl.Path(filename).stat().st_size
        except AssertionError:
            print(f"ERROR : The file {filename} is an empty file")
            raise

    def check_aux_filename(self, aux_filename):

        try:
            assert pl.Path(aux_filename).is_file()
        except AssertionError:
            print(f"ERROR : The auxiliary file {aux_filename} is not a valid file")
            raise

        try:
            assert pl.Path(aux_filename).suffix == '.nc'
        except AssertionError:
            print(f"ERROR : The auxiliary file {aux_filename} does not have a .nc filename ending")
            raise

        try:
            assert pl.Path(aux_filename).stat().st_size
        except AssertionError:
            print(f"ERROR : The auxiliary file {aux_filename} is an empty file")
            raise

    def check_inst_name(self, spif_filename, inst_name):

        dset = nc.Dataset(spif_filename, mode = 'r', format = 'NETCDF4')

        groups = [x for x in dset.groups.keys()]
        dset.close()
        try:
            assert inst_name in groups
        except AssertionError:
            print(f"ERROR : The provided instrument name {inst_name} is not in the groups" + \
                    f" of the file {spif_filename} : {' | '.join(groups)}")
            raise

    def check_config_filename(self, config_filename):

        try:
            open(config_filename, 'r')
        except OSError:
            print(f"ERROR : The provided output filename {config_filename} is not a valid filename.")
            raise

        try:
            assert pl.Path(config_filename).suffix == '.ini'
        except AssertionError:
            print(f"ERROR : The provided filename {config_filename} does not have a .ini file ending")
            raise

    def check_output_filename(self, output_filename):

        try:
            open(output_filename, 'w')
        except OSError:
            print(f"ERROR : The output file argument {output_filename} is not a valid file")
            raise

        try:
            assert pl.Path(output_filename).suffix == '.nc'
        except AssertionError:
            print(f"ERROR : The output file argument {output_filename} does not have a .nc filename ending")
            raise

class AuxFileChecker:

    def __init__(self, aux_filename, config_filename) -> None:
        self.aux_filename = aux_filename
        self.config_filename = config_filename
        self.config = self.get_config()
        self.aux_var_names = self.get_aux_var_names()

    def get_config(self):
        config = configparser.ConfigParser(allow_no_value=True)
        config.read(self.config_filename)

        return config

    def get_aux_var_names(self):
        dset = nc.Dataset(self.aux_filename, mode = 'r', format = 'NETCDF4')
        aux_varnames = [x for x in dset.variables.keys()]
        dset.close()
        return aux_varnames

    def check_aux_file(self):
        self.check_aux_file_open()
        self.check_tas()
        self.check_data_disc()

    def check_aux_file_open(self):
        try:
            dset = nc.Dataset(self.aux_filename, mode = 'r', format = 'NETCDF4')
            dset.close()
        except Exception:
            print(f"The auxiliary file {self.aux_filename} could not be opened for reading.")
            raise

    def check_tas(self):
        orig_tas_name = self.config['aux_data'].get('orig_tas', None)
        corr_tas_name = self.config['aux_data'].get('corrected_tas', None)

        try:
            assert orig_tas_name in self.aux_var_names
        except AssertionError:
            print(f"The supplied original TAS name {orig_tas_name} could not be found in the auxiliary file.")
            raise

        try:
            assert corr_tas_name in self.aux_var_names
        except AssertionError:
            print(f"The supplied corrected TAS name {corr_tas_name} could not be found in the auxiliary file.")
            raise

    def check_data_disc(self):

        for k,v in self.config['data_disc'].items():

            try:
                assert v in self.aux_var_names
            except AssertionError:
                print(f"ERROR : The variable {v} for auxiliary parameter {k} is not present in" + \
                      f" the auxiliary file {self.aux_filename}.")
                raise


class ArgsTransformer:

     def __init__(self, args) -> None:
         self.args = args

         self.transformed_args = {
              'spif_filename':None,
              'aux_filename':None,
              'config':None,
              'inst_names':None,
              'output_filename':None
         }

     def transform_args(self):
          self.transform_filename()
          self.transform_aux_filename()
          self.transform_config()
          self.transform_inst_names()
          self.transform_output_filename()

          return self.transformed_args

     def transform_filename(self):
          self.transformed_args['spif_filename'] = pl.Path(self.args.spif_filename)

     def transform_aux_filename(self):
          self.transformed_args['aux_filename'] = pl.Path(self.args.aux_filename)

     def transform_config(self):
          self.transformed_args['config_filename'] = self.args.config_filename

     def transform_inst_names(self):
        dset = nc.Dataset(self.args.spif_filename, mode = 'r', format = 'NETCDF4')

        if self.args.inst_names == 'all':
            inst_names = [k for k in dset.groups.keys()]
        else:
            inst_names = [self.args.inst_names]

        dset.close()

        self.transformed_args['inst_names'] = inst_names

     def transform_output_filename(self):
        if self.args.output_filename == None:
            self.transformed_args['output_filename'] = pl.Path(self.args.spif_filename)
        else:
            self.transformed_args['output_filename'] = pl.Path(self.args.output_filename)


def add_aux_to_spif(spif_filename, aux_file, inst_names, config_file, output_filename):
    """ Function which adds auxiliary data to SPIF NetCDF file from L2
    atmospheric state NetCDF file.

    Parameters
    ----------
    spif_filename : str or Dataset
        Filename or NetCDF dataset of SPIF NetCDF file to modify.
    aux_file : str
        Filename of L2 atmospheric state NetCDF file to take aux data from.
    inst_name : str
        Instrument name to modify in SPIF file.
    config_file : str, optional
        Name of config file containing parameter definitions to copy to SPIF
        file. If provided, other optional arguments are ignored.
        If not provided, orig_tas must be defined.
    output_filename : str, optional
        If provided, SPIF file is copied to specified directory and aux data
        written in new file.
    """

    spif_rootgrp = nc.Dataset(spif_filename, 'r+')

    # Read parameters from config_file

    config = configparser.ConfigParser()
    config.read(config_file)

    aux_config = config['aux_data']
    orig_tas = aux_config['orig_tas']
    corrected_tas = aux_config.get('corrected_tas', None)

    data_disc = config['data_disc']

    for instrument in inst_names:

        print(f'\nAdding aux to {instrument}...')

        spif_instgrp = spif_rootgrp[instrument]

        if 'aux' not in spif_instgrp.groups:
            spif_instgrp.createGroup('aux')
        spif_auxgrp = spif_instgrp['aux']

        # Get start_date from SPIF File to use in recalculating aux timestamps.
        spif_startdate = datetime.datetime.strptime(
            spif_rootgrp.start_date.strip(), '%Y-%m-%d %H:%M:%S')

        # Open aux file.
        aux_rootgrp = nc.Dataset(aux_file)
        aux_time = aux_rootgrp['Time']

        aux_rootgrp.set_auto_mask(False)

        # Get datetime objects for timestamps in aux_file.
        aux_datetime = convert_netcdf_time(aux_time[:],
                                           aux_time.units)

        # Create AuxTime dimension in SPIF file if it does not exist.
        if 'AuxTime' not in spif_auxgrp.dimensions:
            spif_auxgrp.createDimension(
                'AuxTime', aux_rootgrp.dimensions['Time'].size)

        # Write aux timestamps to SPIF file in aux group.
        write_aux_time(spif_auxgrp, aux_datetime, spif_startdate)

        # Copy original TAS variable to SPIF file.
        copy_aux_variable(spif_auxgrp,
                          'TAS_original',
                          aux_rootgrp,
                          orig_tas)

        # If corrected_tas provided, copy to SPIF file.
        if corrected_tas is not None:
            copy_aux_variable(spif_auxgrp,
                              'TAS_corrected',
                              aux_rootgrp,
                              corrected_tas)

        # If aux_params are provided, copy to SPIF file.
        for dst_name, src_name in data_disc.items():
            copy_aux_variable(spif_auxgrp,
                              dst_name,
                              aux_rootgrp,
                              src_name)

        for key, value in data_disc.items():
                spif_auxgrp.setncattr(key, value)

def write_aux_time(spif_auxgrp, aux_datetime, spif_startdate):
    """ Writes aux timestamps to SPIF file after converting to use same
    reference time.

    Parameters
    ----------
    spif_auxgrp : netCDF group
        Group object pointing to aux group in SPIF NetCDF
    aux_datetime : array of datetime
        Datetimes corresponding to data in aux file.
    spif_startdate : str
        String start_date of SPIF data.
    """
    spif_aux_seconds = nc.date2num(aux_datetime, f'seconds since {spif_startdate}')

    if 'aux_seconds' not in spif_auxgrp.variables:
        fv = numpy.NaN

        zlib = True
        complevel = 4
        out = spif_auxgrp.createVariable('aux_seconds',
                                         'f8',
                                         ('AuxTime',),
                                         fill_value=fv,
                                         zlib=zlib,
                                         complevel=complevel)
    else:
        out = spif_auxgrp['aux_seconds']

    attrs = {'standard_name': 'time',
             'units': 'seconds since start_date',
             'strptime_format': 'seconds since %F %T %z',
             'long_name': 'auxiliary data timestamp'}

    out.setncatts(attrs)
    out[:] = spif_aux_seconds


def copy_aux_variable(spif_auxgrp, spif_aux_var_name, aux_grp, aux_var_name,
    dims=('AuxTime',)):
    """ Given names for source and destination NetCDF groups and variable
    names, copies data from source to destination (creating variables as
    needed).

    Parameters
    ----------
    spif_auxgrp : NetCDF group
        Group object pointing to aux group in SPIF NetCDF
    spif_aux_var_name : str
        Variable name to use for aux data in SPIF file
    aux_grp : NetCDF group
        Group object pointing to source group in aux NetCDF
    aux_var_name : str
        Name of source variable
    dims : tuple, optional
        Dimensions to use for destination variable. Defaults to ('AuxTime',).
    """

    var = aux_grp[aux_var_name]

    if spif_aux_var_name not in spif_auxgrp.variables:
        fv = None
        if '_FillValue' in var.ncattrs():
            fv = var.getncattr('_FillValue')

        zlib = True
        complevel = 4

        out = spif_auxgrp.createVariable(spif_aux_var_name,
                                         var.dtype,
                                         dims,
                                         fill_value=fv,
                                         zlib=zlib,
                                         complevel=complevel,
                                         )
    else:
        out = spif_auxgrp[spif_aux_var_name]

    out_dict = var.__dict__
    out_dict.pop('_FillValue')

    if spif_aux_var_name != aux_var_name:
        out_dict['source_variable'] = aux_var_name

    out.setncatts(out_dict)
    out.setncattr('AuxSource', aux_grp.filepath())

    var_data = var[:]
    out[:] = var_data


def convert_netcdf_time(time, time_basis, start_date=None):
    """ Converts time in standard NetCDF time format (seconds since given
    date-time) into datetime object.

    Parameters
    ----------
    time : array
        Timestamps of data contained within the file. Seconds since
        specified date-time
    time_basis : str
        Format of timestamps, e.g., 'seconds since 2020-07-14 00:00:00'
    start_date : str, optional
        String start date to ues for time_basis, in the event time basis
        is defined in reference to this date.

    Returns
    -------
    list of datetime object
        Absolute datetimes of provided times.
    """

    if start_date is not None:
        time_basis = time_basis.replace('start_date', start_date)

    time_dt = nc.num2date(time, time_basis)

    return time_dt



#!/usr/bin/env python
# coding: utf-8

import gc
import time

import numpy

from .utils import list_to_array, list3d_to_array, _reshape_image


class Images(object):
    """ Class containing info for set of images. Default info includes
    nanoseconds, seconds, image, length, and buffer index. Users can add other
    parameters by passing list at initialization with names of other parameters
    to store.

    Class Attributes
    ----------------
    default_items : list of str
        List of default items common to all image sources.

    Attributes
    ----------
    ns : list
        Nanoseconds timestamp for images.
    sec : list
        Seconds timestamp for images.
    image : list
        Image data for images.
    length : list
        Length data, in slices, for images.
    buffer_index : list
        Buffer index value for images.

    Parameters
    ----------
    aux_names : list of str, optional
        List of strings of auxiliary parameters to include in current Images
        instance. If provided, strings in list become class attributes.
    """
    default_items = ['ns', 'sec', 'image', 'length', 'buffer_index']

    def __init__(self, aux_names=None):
        for item in self.default_items:
            setattr(self, item, [])
        if aux_names is not None:
            for name in aux_names:
                setattr(self, name, [])

    def __len__(self):
        return len(self.sec)

    def clear(self):
        """ Resets all class attributes to empty lists.
        """
        for key in self.__dict__.keys():
            setattr(self, key, [])
        gc.collect()

    def extend(self, p):
        """ Extends all attributes in current instance with corresponding
        attributes in given instance.

        Parameters
        ----------
        p : Images object
            Images object to copy data from to current instance.
        """
        if p is not None:
            if len(p) > 0:
                for key, val in self.__dict__.items():
                    getattr(self, key).extend(getattr(p, key))

    def reshape_image(self, diodes):
        """ Reshapes image data to 3-D array, with dimensions of image, slices,
        and diodes. Should be called only before passing to write function, as
        reshaped images cannot be extended with other image data.

        Parameters
        ----------
        diodes : int
            Number of diodes of current instrument.
        """
        t0 = time.time()
        image, max_slices = list_to_array(self.image, diodes)
        print(time.time() - t0)
        t0 = time.time()
        #self.image = numpy.reshape(image, (-1, max_slices // diodes, diodes))
        self.image = _reshape_image(image, max_slices, diodes)
        print(time.time() - t0)

    def conv_to_array(self, diodes):
        """ Converts list of 2-D arrays to 1-D numpy array. Should be called
        only before passing to write function, as reshaped images cannot be
        extended with other image data.

        Parameters
        ----------
        diodes : int
            Number of diodes of current instrument.
        """
        # self.image = list(itertools.chain.from_iterable(self.image))
        flatten = numpy.concatenate(self.image)
        self.image = flatten.ravel()



#!/usr/bin/env python
# coding: utf-8

from concurrent.futures import FIRST_COMPLETED
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import wait
import datetime
import math
import os
import time

import numpy
from tqdm import tqdm

from . import BinaryFile
from .input_utils import read_sea_image_data, read_sea_tas
from ..images import Images
from ..utils import convert_uint32_to_bitlist

MAX_PROCESSORS = 20


class TwoDFile(BinaryFile):

    syncword = int('0x55000000', 16)

    def __init__(self, filename, inst_name, resolution):
        super().__init__(filename, inst_name, resolution)
        self.diodes = 32
        self.file_dtype = numpy.dtype([('year', 'u2'),
                                       ('month', 'u2'),
                                       ('day', 'u2'),
                                       ('hour', 'u2'),
                                       ('minute', 'u2'),
                                       ('second', 'u2'),
                                       ('ms', 'u2'),
                                       ('weekday', 'u2'),
                                       ('data', '(1024, )u4')
                                       ])
        self.aux_channels = ['tas', 'timeword']

    def read(self):

        self.tas = read_sea_tas(self.filename, '2dc', self.resolution)
        self.data = read_sea_image_data(self.filename, self.file_dtype, '2d', self.name)

        self.get_start_date()
        self.calc_buffer_datetimes()

    def process_file(self, spiffile, processors=None):
        """ Method to process file using multiple processors. Calls
        process_image method implemented in children classes. Passes image
        data to spiffile object to be written to file.

        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        processors : int, optional
            Number of processors to use for parallel image processing. If none
            is provided, defaults to number of processors minus one.
        """

        if processors is None:
            processors = os.cpu_count()
            if os.cpu_count() > 1:
                processors -= 1

        if processors > MAX_PROCESSORS:
            processors = MAX_PROCESSORS

        spiffile.set_start_date(self.start_date.strftime('%Y-%m-%d %H:%M:%S %z'))

        process_until = len(self.data)
        data_chunk = range(0, process_until)

        tot_images = 0

        spiffile.create_inst_group(self.name)
        spiffile.write_buffer_info(self.start_date, self.datetimes)
        spiffile.set_filenames_attr(self.name, self.filename)

        pbar1 = tqdm(desc='Processing frames',
                     total=process_until,
                     unit='frame')
        pbar2 = tqdm(desc='Writing frames',
                     total=process_until,
                     unit='frame')
        t00 = time.time()

        i = 0
        chunksize = 500
        max_write_queue = 8
        images_remaining = True
        futures = []
        with ProcessPoolExecutor(max_workers=processors) as executor:
            while True:
                while len(futures) <= max_write_queue and images_remaining:
                    futures.append(executor.submit(self.process_frames,
                                                   data_chunk[i: i + chunksize]))
                    i += chunksize
                    if i >= process_until:
                        images_remaining = False
                    pbar1.update(chunksize)

                done, running = wait(futures, return_when=FIRST_COMPLETED)
                for f in done:
                    indx = futures.index(f)
                    if indx == 0:
                        pbar2.update(chunksize)
                        images = f.result()
                        tot_images += len(images)
                        t0 = time.time()
                        if len(images) > 0:
                            images.conv_to_array(self.diodes)
                            spiffile.write_images(self.name, images)

                        futures.pop(indx)

                if not images_remaining and len(futures) == 0:
                    break

        pbar1.close()
        pbar2.close()
        print('Finished processing.')
        t11 = time.time()
        print(f'{tot_images} images processed in {t11 - t00:0.3f} seconds')

    def process_frames(self, frames):
        p = Images(self.aux_channels)

        for frame in frames:
            images = self.process_frame(frame)
            p.extend(images)

        return p

    def process_frame(self, frame):
        """ Method to process single frame of image data. Decompresses image
        data from frame and passes resulting image data to extract_images
        method to extract image info from image buffer.

        Parameters
        ----------
        frame : int
            Frame number in self.data to process.

        Returns
        -------
        Images object
            Images object containing image info in current frame.
        """
        data = self.data[frame]
        record = list(data['data'])
        tas = self.tas[frame]

        record_time = datetime.datetime(data['year'],
                                        data['month'],
                                        data['day'],
                                        data['hour'],
                                        data['minute'],
                                        data['second'],
                                        data['ms'] * 1000)
        # if frame % 1000 == 0:
        #     print('At frame ' + str(frame) + ' of ' + str(len(self.data)))
        return self.extract_images(frame, record, record_time, tas)

    def extract_images(self, frame, frame_decomp, record_time, tas):
        """ Extracts images and associated info from decompressed image
        buffer. The image image, length, timestamp in sec and ns, image
        count, dof flag, and buffer index are extracted for each image.

        Parameters
        ----------
        frame : int
            Frame number of current frame being processed.
        frame_decomp : array
            Array of decompressed image buffer.
        data : datetime object
            Date of current image buffer.

        Returns
        -------
        Images object
            Images object containing image data extracted from image
            buffer.
        """
        images = Images(self.aux_channels)
        frequency = self.resolution * 1e-6 / tas

        ii = 0

        syncwords = []
        while ii < len(frame_decomp):
            ii = self.search_syncword(frame_decomp, ii)
            syncwords.append(ii)
            ii += 1

        ref_timeword = None
        timeword_rollovers = 0
        for i, syncword in enumerate(syncwords[:-1]):
            next_syncword = syncwords[i + 1]
            im = frame_decomp[syncword + 1: next_syncword - 1]
            timeword = frame_decomp[next_syncword - 1] & int('0x00ffffff', 16)

            if ref_timeword is None:
                ref_timeword = timeword
                prev_timeword = timeword

            if timeword - prev_timeword < 0:
                timeword_rollovers += 1

            delta_timeword = timeword - ref_timeword \
                + timeword_rollovers * int('0x00ffffff', 16)

            delta_time = delta_timeword * frequency

            image_time = record_time + datetime.timedelta(seconds=delta_time)

            epoch_time = (image_time - self.start_date).total_seconds()
            ns, sec = math.modf(epoch_time)

            image = numpy.array(convert_uint32_to_bitlist(im))
            image = image.flatten()

            if len(image) > 0:
                images.ns.append(ns * 1e9)
                images.sec.append(sec)
                images.length.append(len(im))
                images.image.append(image)
                images.tas.append(tas)
                images.timeword.append(timeword)
                images.buffer_index.append(frame)

            prev_timeword = timeword

        return images

    def search_syncword(self, frame_decomp, i):
        """ Searches image buffer for next image using syncword. If not found
        returns index of end of current image buffer.

        Parameters
        ----------
        frame_decomp : array
            Array of decompressed image buffer.
        i : int
            Index in frame_decomp to search from.

        Returns
        -------
        int
            Index value of start of next syncword.
        """
        try:
            while True:
                i = frame_decomp.index(self.syncword, i)
                if (frame_decomp[i - 1] != int('0xffffffff', 16)) & \
                    (frame_decomp[i - 2] == int('0xffffffff', 16)):
                    return i
                i += 1
        except (ValueError, AttributeError, IndexError) as e:
            return len(frame_decomp)

    def calc_buffer_datetimes(self):
        """ Calculates datetimes from bufffers read in from file and sets
        to datetimes class attribute.
        """
        self.datetimes = [datetime.datetime(d['year'],
                                            d['month'],
                                            d['day'],
                                            d['hour'],
                                            d['minute'],
                                            d['second'],
                                            d['ms'] * 1000) for d in self.data]
